﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Web.Services;
using System.Xml;
using System.Xml.Schema;
using System.IO;

/// <summary>
/// This is a sample service provided by the State of Alabama to help
/// companies implement the requirements of the Online Insurance Verification System (OIVS)
/// The service and its namespace is generic, so it could be used for other XML transmissions
/// and will easily support other schemas in the future. The service takes in a string of XML
/// along with the schema version of that xml and returns back the appropriate XML based on that
/// request. Enhancements such as checking that the requested schema exist on disk and error 
/// checking and handling should be included but for the purpose of this demo solution 
/// were omitted due to time constraints. Also, SOAP 1.2 is suppressed in the web.config file
/// since the IICMVA specification that this service targets required SOAP 1.1 only.
/// </summary>
[WebService(Namespace = "http://xmlexchangeservice.com/",
   Description = "XML transefer web service to support multiple data exchanges.")]
[WebServiceBinding(ConformsTo = WsiProfiles.BasicProfile1_1)]
public class XMLExchangeService : System.Web.Services.WebService
{
    public XMLExchangeService(){}

    private string output = "";

    [WebMethod(Description = "Transmits XML to the service and the corresponding scema version. The service returns XML corresponding to the input request. Security is handled at the web server level via client certificates.")]
    public string TransmitXML(string InputXML, string XMLSchemaVersion)
    {
        //****IMPORTANT NOTES TO DEVELOPERS****
            //NOTE 1:
                //Both the 2008 and the 2005 IICMVA schemas technically support multiple
                //verifications per transaction but Alabama will not be requiring companies to 
                //implement this functionality. These 'Bulk' queries will not me used and a
                //developer can code against ONE 'Detail' XML section per transaction and will
                //only need to respond with ONE 'Detail' XML section in the response.
            //NOTE 2:
                //'PolicyKey' (2008 schema) and 'UniqueKey' (2005 schema) WILL NOT contain spaces, hyphens, etc.
                //These values will be passed to you as ALPHA NUMERIC ONLY and you will need to match
                //them accordingly to your data.
            //NOTE 3:
                //Alabama will be supporting multiple schemas (00200510 and 00200809 specifially)
                //so we test the incoming query and determine the schema type and then test to 
                //make sure that it is conformant to the requested schema. For a company specific 
                //service, the code in either if/else is recommended as the 'contains' statement
                //will not be needed nor the 'other' schema check as a particular company
                //will only be supporting one schema.
            //NOTE 4:
                //It was determined that the security for this service would happen as follows:
                //This service would be hosted on a web server that restricts IP access to the IP
                //of the calling Alabama server. In addition, the server would have the public certificate
                //that Alabama has the corresponding certificate for and access would be limited based on
                //that certificate relationship. More information on configuring IIS7 with client certificate
                //mappings can be found here: http://www.iis.net/ConfigReference/system.webServer/security/authentication/iisClientCertificateMappingAuthentication
        //****END IMPORTANT NOTES TO DEVELOPERS****


 
        //THIS SECTION IS FOR 2005 SCHEMA COMPANIES.
        //REMOVE THIS SECTION IF YOU ARE NOT A 2005
        //SCHEMA COMPANY
        
        //****BEGIN 2005 SCHEMA SECTION****

        //Test if the schema is 2005
        if(XMLSchemaVersion == "urn-schemas-x12-org-00200510-CoverageRequest")
        {

            //Validate the schema
            output = ValidateSchema(InputXML, "schema\\" + XMLSchemaVersion + ".xsd");

            //****INSERT YOUR REAL LOGIC HERE****
            //The output from the schema validation above will need to be checked for errors. If no errors,
            //a company would then parse the XML and extract the VIN, Policy Number, and Verification Date.
            //They would then verify that information against their internal database and return back a response
            //that is conformant to the appropriate XML schema that they are using (2005 in the case of this code section).
            //Appropriate reason codes should be supplied for 'Unconfirmed' responses.

            //BEGIN SAMPLE FAKE CODE
            if (output == "")
            {
                //The code below is only for testing/sample purposes. A real service would have parsed the input XML, created
                //a query to a company DB and returned the result in XML conformant to the schema used.
                //The schema of the response is in the 'schema' folder of this application and should be used
                //to create the XML response. A response that is not conformant to the schema indicated by
                //the company will be rejected by the calling service.
                try
                {
                    InsuranceValidationRequest request = ParseRequest(InputXML, "schema\\" + XMLSchemaVersion + ".xsd", XMLSchemaVersion);
                    if (request != null)
                    {
                        //Run the data against a company DB

                        //Return the appropriate resonse
                        output = Create2005Response("Unconfirmed", "8", request.NAIC, request.VerificationDate, request.PolicyNumber, request.PolicyState);
                    }
                    else
                    {
                        //return an apporpriate response
                    }
                }
                catch (Exception ex)
                {
                    //return an apporpriate response
                }
            }
            //END SAMPLE FAKE CODE
        }
        //****END 2005 SCHEMA SECTION****


        //THIS SECTION IS FOR 2008 SCHEMA COMPANIES.
        //REMOVE THIS SECTION IF YOU ARE NOT A 2008
        //SCHEMA COMPANY

        //****BEGIN 2008 SCHEMA SECTION****

         //Test if the schema is 2008
        else if(XMLSchemaVersion == "urn-schemas-x12-org-V00200806-CoverageRequest")
        {

            //Validate the schema
            output = ValidateSchema(InputXML, "schema\\" + XMLSchemaVersion + ".xsd");

            //****INSERT YOUR REAL LOGIC HERE****
            //The output from the schema validation above will need to be checked for errors. Assuming no errors,
            //a company would then parse the XML and extract the VIN, Policy Number, and Verification Date.
            //They would then verify that information against their internal database and return back a response
            //that is conformant to the appropriate XML schema that they are using (2008 in the case of this code section)
            //Appropriate reason codes should be supplied for 'Unconfirmed' responses.

            //BEGIN SAMPLE FAKE CODE
            if (output == "")
            {
                //The code below is only for testing/sample purposes. A real service would have parsed the input XML, created
                //a query to a company DB and returned the result in XML conformant to the schema used.
                //The schema of the response is in the 'schema' folder of this application and should be used
                //to create the XML response. A response that is not conformant to the schema indicated by
                //the company will be rejected by the calling service.
                try
                {
                    InsuranceValidationRequest request = ParseRequest(InputXML, "schema\\" + XMLSchemaVersion + ".xsd", XMLSchemaVersion);
                    if (request != null)
                    {
                        //Run the data against a company DB

                        //Return the appropriate resonse
                        output = Create2008Response("Unconfirmed", "PKEY3", request.NAIC, request.VerificationDate, request.PolicyNumber, request.PolicyState);
                    }
                    else
                    {
                        //return an apporpriate response
                    }
                }
                catch (Exception ex)
                {
                    //return an apporpriate response
                }
            }
            //END SAMPLE FAKE CODE
        }
         //****END 2008 SCHEMA SECTION****

        else
        {
            //Invalid XML schema request. Technically both schemas support an invalid XML request. Ideally the appropriately
            //formatted response should be returned that is conformant to the schema that a particular company is using.
            //Sample invalid query responses are included in the schema folder and those are what should be used for an
            //invalid (non-schema conformant) query.

            //THIS EXAMPLE IS JUST A POSSIBLE DATA FLOW. In reality, 
            
            //I picked the 2008 invlid response for this sample, but a company would respond with the appropriate schema.
            string xmlPath = System.Web.HttpContext.Current.Server.MapPath("schema\\2008InvalidQueryResponse.xml");
            output = File.ReadAllText(xmlPath);
        }

        return output;
    }

    private string Create2008Response(string response, string responseReason, string naic, DateTime verificationDate, string policyNumber, string policyState)
    {
        //Sample code only

        StringWriter sw = new StringWriter();
        string retVal = "";
        XmlTextWriter writer = new XmlTextWriter(sw);
        writer.Formatting = Formatting.Indented;
        writer.WriteStartDocument();
        writer.WriteStartElement(null, "tns:CoverageResponseDocument", null);
        writer.WriteAttributeString(null, "xmlns:X12.7", null, "urn:x12:schemas:V00200806:X12.7");
        writer.WriteAttributeString(null, "xmlns:tns", null, "urn:schemas:x12:org:V00200806:CoverageResponseDocument");
        writer.WriteAttributeString(null, "xmlns:xsi", null, "http://www.w3.org/2001/XMLSchema-instance");
        writer.WriteAttributeString(null, "xsi:schemaLocation", null, "urn:schemas:x12:org:V00200806:CoverageResponseDocument CoverageResponse-00200809.xsd");
        writer.WriteAttributeString(null, "PublicationVersion", null, "00200809");
        writer.WriteAttributeString(null, "PublicationDate", null, "1980-11-02");

        writer.WriteStartElement(null, "tns:Detail", null);
        writer.WriteStartElement(null, "tns:PolicyInformation", null);
        writer.WriteStartElement(null, "tns:CoverageStatus", null);
        writer.WriteStartElement(null, "tns:ResponseDetails", null);
        writer.WriteStartElement(null, "tns:ResponseCode", null);
        writer.WriteString(response);
        writer.WriteEndElement();
        writer.WriteStartElement(null, "tns:UnconfirmedReasonCode", null); //Only used if the response is Unconfirmed
        writer.WriteString(responseReason);
        writer.WriteEndElement();
        writer.WriteEndElement();
        writer.WriteEndElement();
        writer.WriteStartElement(null, "tns:OrganizationDetails", null);
        writer.WriteStartElement(null, "tns:NAIC", null);
        writer.WriteString(naic);
        writer.WriteEndElement();
        writer.WriteEndElement();
        writer.WriteStartElement(null, "tns:PolicyDetails", null);
        writer.WriteStartElement(null, "tns:VerificationDate", null);
        writer.WriteString(verificationDate.ToString("yyyy-MM-dd'T'HH:mm:ss.fffffffZ"));
        writer.WriteEndElement();
        writer.WriteStartElement(null, "tns:PolicyKey", null);
        writer.WriteString(policyNumber);
        writer.WriteEndElement();
        writer.WriteStartElement(null, "tns:PolicyState", null);
        writer.WriteString(policyState);
        writer.WriteEndElement();
        writer.WriteEndElement();
        writer.WriteEndElement();
        writer.WriteEndElement();

        writer.WriteEndElement();
        writer.Close();
        retVal = sw.ToString();

        return retVal;
    }

    private string Create2005Response(string response, string responseReason, string naic, DateTime verificationDate, string policyNumber, string policyState)
    {
        //Sample code only

        StringWriter sw = new StringWriter();
        string retVal = "";
        XmlTextWriter writer = new XmlTextWriter(sw);
        writer.Formatting = Formatting.Indented;
        writer.WriteStartDocument();
        writer.WriteStartElement(null, "tns:CoverageResponse", null);
        writer.WriteAttributeString(null, "xmlns:tns", null, "urn:schemas:x12:org:00200510:CoverageResponse");
        writer.WriteAttributeString(null, "xmlns:xsi", null, "http://www.w3.org/2001/XMLSchema-instance");
        writer.WriteAttributeString(null, "xsi:schemaLocation", null, "urn:schemas:x12:org:00200510:CoverageResponse CoverageResponse-00200510.xsd");

        writer.WriteStartElement(null, "Detail", null);
        writer.WriteStartElement(null, "PolicyInformation", null);
        writer.WriteStartElement(null, "CoverageStatus", null);
        writer.WriteStartElement(null, "ResponseDetails", null);
        writer.WriteStartElement(null, "ResponseCode", null);
        writer.WriteString(response);
        writer.WriteEndElement();
        writer.WriteStartElement(null, "UnconfirmedReasonCode", null); //Only used if the response is Unconfirmed
        writer.WriteString(responseReason);
        writer.WriteEndElement();
        writer.WriteEndElement();
        writer.WriteEndElement();
        writer.WriteStartElement(null, "OrganizationDetails", null);
        writer.WriteStartElement(null, "NAIC", null);
        writer.WriteString(naic);
        writer.WriteEndElement();
        writer.WriteEndElement();
        writer.WriteStartElement(null, "PolicyDetails", null);
        writer.WriteStartElement(null, "VerificationDate", null);
        writer.WriteString(verificationDate.ToString("yyyy-MM-dd'T'HH:mm:ss.fffffffZ"));
        writer.WriteEndElement();
        writer.WriteStartElement(null, "UniqueKey", null);
        writer.WriteString(policyNumber);
        writer.WriteEndElement();
        writer.WriteStartElement(null, "PolicyState", null);
        writer.WriteString(policyState);
        writer.WriteEndElement();
        writer.WriteEndElement();
        writer.WriteEndElement();
        writer.WriteEndElement();

        writer.WriteEndElement();
        writer.Close();
        retVal = sw.ToString();

        return retVal;
    }

    private InsuranceValidationRequest ParseRequest(string inputXML, string pathToXSD, string XMLSchemaVersion)
    {
        InsuranceValidationRequest retVal = new InsuranceValidationRequest();
        string output = "";
        XmlSchemaSet sc = new XmlSchemaSet();
        string xsdPath = System.Web.HttpContext.Current.Server.MapPath(pathToXSD);
        XmlTextReader tr = new XmlTextReader(xsdPath);
        sc.Add(null, tr);

        // Set the validation settings.
        XmlReaderSettings settings = new XmlReaderSettings();
        settings.ValidationType = ValidationType.Schema;
        settings.Schemas = sc;
        settings.ValidationEventHandler += new ValidationEventHandler(ValidationCallBack);

        // Create the XmlReader object.
        XmlReader reader = XmlReader.Create(new StringReader(inputXML), settings);

        //Alabama supports 2 schemas, a company would only need 1 of the following:
        if (XMLSchemaVersion == "urn-schemas-x12-org-00200510-CoverageRequest")
        {
            try
            {
                // Parse the file. 
                while (reader.Read())
                {
                    if (reader.NodeType == XmlNodeType.Element)
                    {
                        switch (reader.Name)
                        {
                            case "Name":
                                reader.Read();
                                if (reader.NodeType == XmlNodeType.Text)
                                {
                                    retVal.OrganizationName = reader.Value;
                                }
                                break;
                            case "ReasonCode":
                                reader.Read();
                                if (reader.NodeType == XmlNodeType.Text)
                                {
                                    retVal.QueryReason = reader.Value;
                                }
                                break;
                            case "NAIC":
                                reader.Read();
                                if (reader.NodeType == XmlNodeType.Text)
                                {
                                    retVal.NAIC = reader.Value;
                                }
                                break;
                            case "VerificationDate":
                                reader.Read();
                                if (reader.NodeType == XmlNodeType.Text)
                                {
                                    retVal.VerificationDate = Convert.ToDateTime(reader.Value);
                                }
                                break;
                            case "UniqueKey":
                                reader.Read();
                                if (reader.NodeType == XmlNodeType.Text)
                                {
                                    retVal.PolicyNumber = reader.Value;
                                }
                                break;
                            case "PolicyState":
                                reader.Read();
                                if (reader.NodeType == XmlNodeType.Text)
                                {
                                    retVal.PolicyState = reader.Value;
                                }
                                break;
                            case "VIN":
                                reader.Read();
                                if (reader.NodeType == XmlNodeType.Text)
                                {
                                    retVal.VIN = reader.Value;
                                }
                                break;
                        }
                    }
                }
            }
            catch (Exception ex)
            {

                return null;
            }
       }
        else if (XMLSchemaVersion == "urn-schemas-x12-org-V00200806-CoverageRequest")
        {
            try
            {
                // Parse the file. 
                while (reader.Read())
                {
                    if (reader.NodeType == XmlNodeType.Element)
                    {
                        switch (reader.Name)
                        {
                            case "tns:Name":
                                reader.Read();
                                if (reader.NodeType == XmlNodeType.Text)
                                {
                                    retVal.OrganizationName = reader.Value;
                                }
                                break;
                            case "tns:ReasonCode":
                                reader.Read();
                                if (reader.NodeType == XmlNodeType.Text)
                                {
                                    retVal.QueryReason = reader.Value;
                                }
                                break;
                            case "tns:NAIC":
                                reader.Read();
                                if (reader.NodeType == XmlNodeType.Text)
                                {
                                    retVal.NAIC = reader.Value;
                                }
                                break;
                            case "tns:VerificationDate":
                                reader.Read();
                                if (reader.NodeType == XmlNodeType.Text)
                                {
                                    retVal.VerificationDate = Convert.ToDateTime(reader.Value);
                                }
                                break;
                            case "tns:PolicyKey":
                                reader.Read();
                                if (reader.NodeType == XmlNodeType.Text)
                                {
                                    retVal.PolicyNumber = reader.Value;
                                }
                                break;
                            case "tns:PolicyState":
                                reader.Read();
                                if (reader.NodeType == XmlNodeType.Text)
                                {
                                    retVal.PolicyState = reader.Value;
                                }
                                break;
                            case "tns:VIN":
                                reader.Read();
                                if (reader.NodeType == XmlNodeType.Text)
                                {
                                    retVal.VIN = reader.Value;
                                }
                                break;
                        }
                    }
                }
            }
            catch (Exception ex)
            {

                return null;
            }
        }
        else
        {
            return null;
        }

        return retVal;
    }

    private void ValidationCallBack(object sender, ValidationEventArgs e)
    {
        output = output + "Validation Error: " + e.Message + Environment.NewLine;
    }

    private string ValidateSchema(string inputXML, string pathToXSD)
    {
        string output = "";
        XmlSchemaSet sc = new XmlSchemaSet();
        string xsdPath = System.Web.HttpContext.Current.Server.MapPath(pathToXSD);
        XmlTextReader tr = new XmlTextReader(xsdPath);
        sc.Add(null, tr);

        // Set the validation settings.
        XmlReaderSettings settings = new XmlReaderSettings();
        settings.ValidationType = ValidationType.Schema;
        settings.Schemas = sc;
        settings.ValidationEventHandler += new ValidationEventHandler(ValidationCallBack);

        // Create the XmlReader object.
        XmlReader reader = XmlReader.Create(new StringReader(inputXML), settings);

        try
        {
            // Parse the file. 
            while (reader.Read()) ;
        }
        catch (Exception ex)
        {
            output = output + "Validation Error: " + ex.ToString() + Environment.NewLine;
        }
        return output;
    }
}
